"""
Base classes for configuration models
"""

import collections
import configparser
import functools
import os
import typing

import pydantic

from ... import __project_name__, errors, utils
from .. import string

import logging  # isort:skip
_log = logging.getLogger(__name__)


class _CreateDefaults:
    """
    Descriptor class that provides a cached instance of the parent class

    The parent class is called without arguments, so it should provide only default values.

    The instance is made immutable by setting :attr:`~.pydantic.BaseModel.model_config` to
    ``{"frozen": True}``.

    If the class has also a method called `get_defaults`, it must provide ``(field_name, cls)``
    tuples. `cls` is called without arguments to create the default instance for the field called
    `field_name`.

    >>> class Foo:
    >>>     defaults: ClassVar = _CreateDefaults()
    >>> Foo.defaults
    <immutable instance of Foo>

    >>> Foo.defaults["foo"] = "bar"
    <Exception: Instance is frozen>
    """

    def __set_name__(self, objtype, name):
        self.attribute_name = name

    def __get__(self, obj, objtype):
        # Create special immutable subclass of `objtype` that provides default values.
        DefaultsSubclass = self._create_defaults_subclass(objtype)

        # Store defaults instance on the same attribute that this _CreateDefaults() instance was
        # stored on. This replaces _CreateDefaults() (which is no longer needed) and makes lookup of
        # `objtype.defaults` as fast as possible.
        setattr(objtype, self.attribute_name, DefaultsSubclass())

        # Do a quick sanity check.
        assert type(objtype.defaults).__name__ == f'{objtype.__name__}Defaults', \
            f'Expected {objtype.__name__}Defaults, not {type(objtype.defaults).__name__}'
        return objtype.defaults

    @classmethod
    def _create_defaults_subclass(cls, objtype):
        clsname = f'{objtype.__name__}Defaults'
        bases = (objtype,)
        attrs = {
            # This makes the subclass immutable. `objtype.model_config` should be automatically
            # copied for our *Defaults subclass, only overloading `frozen=True`.
            'model_config': pydantic.ConfigDict(frozen=True),
            '__annotations__': {},
        }

        try:
            # Get function that provides `(field_name, cls)` pairs from the class this descriptor
            # was created for.
            get_defaults = objtype.get_defaults
        except AttributeError:
            # Default implementation of get_defaults().
            def get_defaults():
                for field_name, field_info in objtype.__pydantic_fields__.items():
                    cls = field_info.annotation
                    yield field_name, cls

        for field_name, typ in get_defaults():
            if isinstance(typ, type) and issubclass(typ, ConfigDictBase):
                # Handle nested key-value map.
                typ_defaults = typ.defaults
                attrs[field_name] = typ_defaults
                attrs['__annotations__'][field_name] = type(typ_defaults)

        return type(clsname, bases, attrs)


class ConfigDictBase(collections.abc.MutableMapping, pydantic.BaseModel):
    """
    :class:`pydantic.BaseModel` that provides dictionary field access
    """

    model_config = pydantic.ConfigDict(
        arbitrary_types_allowed=True,
        extra='forbid',
        validate_assignment=True,
        validate_default=True,
    )

    def __getitem__(self, key):
        # "foo-bar" is synonymous with "foo_bar" and vice versa.
        valid_keys = (
                key,
                key.replace('-', '_'),
        )
        for k in valid_keys:
            try:
                return getattr(self, k)
            except AttributeError:
                pass
        raise KeyError(key)

    def __setitem__(self, key, value):
        try:
            setattr(self, key, value)
        except pydantic.ValidationError as e:
            raise self._value_error(e) from e

    def __delitem__(self, key):
        raise RuntimeError(f'Keys cannot be removed from {type(self).__name__}: {key!r}: {self!r}')

    def _iter_all_fields(self):
        # `__pydantic_fields__` has fields that are specifically defined in the subclass.
        for field_name in self.__pydantic_fields__:
            yield field_name

        # `__pydantic_extra__` has fields that were not specifically defined.
        if getattr(self, '__pydantic_extra__', None):
            for field_name in sorted(self.__pydantic_extra__):
                yield field_name

    def __iter__(self):
        yield from self._iter_all_fields()

    def __len__(self):
        return len(tuple(self._iter_all_fields()))

    defaults: typing.ClassVar = _CreateDefaults()
    """
    Class property that holds an instance of this class with default values

    See :class:`_CreateDefaults`.
    """

    def _value_error(self, exception, path=()):
        """Translate :class:`pydantic.ValidationError` into :class:`~.ConfigValueError`"""

        def extra_forbidden(error):
            if len(error['loc']) == 1:
                return 'No such section'
            else:
                return 'No such option'

        def value_error(error):
            if error['msg'].startswith('Value error, '):
                return error['msg'][len('Value error, '):]
            else:
                return error['msg']

        custom_msgs = {
            'extra_forbidden': extra_forbidden,
            'value_error': value_error,
        }

        def maybe_add_input(error):
            if str(error['input']) in error['msg']:
                return error['msg']
            else:
                return error['msg'] + ': ' + str(error['input'])

        # TODO: When ExceptionGroup (introduced in Python 3.11) is widely available, we can raise
        #       mulitple exceptions here so that multiple invalid values in a config file are
        #       reported at the same time.
        for error in exception.errors():
            _log.debug('Validation error: %r', error)
            get_msg = custom_msgs.get(error['type'])
            if get_msg:
                msg = get_msg(error)
            else:
                msg = maybe_add_input(error)
            return errors.ConfigValueError(msg, path=(*path, *error['loc']))

        raise RuntimeError(f'No error found: {exception!r}') from exception


class _GetSectionNames:
    def __get__(self, obj, objtype):
        return tuple(objtype.__pydantic_fields__)


class ConfigBase(ConfigDictBase):
    """
    Root-level configuration model

    Instances map section names to :class:`SectionBase` subclass instances.

    From a user perspective, a section is a file name without the file extension, e.g. "clients" for
    "~/.config/upsies/clients.ini".
    """

    section_names: typing.ClassVar = _GetSectionNames()
    """Sequence of valid section names"""

    @pydantic.model_validator(mode='after')
    def validate_config(self):
        """
        Pass this :class:`ConfigBase` instance to :meth:`.SectionBase.validate_config` on all
        sections

        :return: the :class:`ConfigBase` instance (i.e. the first argument `self`)
        """
        for section in self.values():
            section.validate_config(self)
        return self

    def __getitem__(self, key):
        try:
            return super().__getitem__(key)
        except KeyError:
            raise errors.UnknownSectionConfigError(key) from None

    def dump(self):
        """Return all values as a human-readable tree structure string"""
        return self._dump(self)

    def dump_defaults(self):
        """Return default values as a human-readable tree structure string"""
        return self._dump(self.defaults)

    @staticmethod
    def _dump(dct):
        lines = []
        for section_name, section in sorted(dct.items()):
            lines.append(f'   {section_name} [{type(section).__name__}]')
            for subsection_name, subsection in sorted(section.items()):
                lines.append(f'      {subsection_name} [{type(subsection).__name__}]')
                for option_name, value in sorted(subsection.items()):
                    lines.append(f'         {option_name} [{type(value).__name__}] = {value!r}')
        return '\n'.join(lines)

    @functools.cached_property
    def files(self):
        """
        :class:`~.collections.abc.Mapping` of section names to file paths

        Section names should be file stems, i.e. file names without extensions.
        """
        return {}

    def read(self, section_name, filepath, *, ignore_missing=False):
        """
        Read `filepath` and make its contents available as `section`

        :raises ConfigError: if reading, parsing or validating `filepath` fails
        """
        _log.debug('Loading %s configuration from %s', section_name, filepath)
        section_kwargs = self._read_ini(filepath, ignore_missing=ignore_missing)
        section_cls = self.__pydantic_fields__[section_name].annotation
        try:
            section = section_cls(**section_kwargs)
            setattr(self, section_name, section)
        except pydantic.ValidationError as e:
            raise errors.ConfigError(f'{filepath}: {self._value_error(e)}') from e
        else:
            # Remember the origin of the config so we can write() it later.
            self.files[section_name] = filepath

    def _read_ini(self, filepath, *, ignore_missing=False):
        # Create INI parser.
        cfgparser = configparser.ConfigParser(
            default_section=None,
            interpolation=None,
        )

        # Read INI file if it exists.
        filecontent = self._read_file(filepath, ignore_missing=ignore_missing)
        if filecontent:
            # Parse raw INI string.
            try:
                cfgparser.read_string(filecontent, source=filepath)
            except configparser.MissingSectionHeaderError as e:
                raise errors.ConfigError(f'{filepath}: Line {e.lineno}: {e.line.strip()}: Option outside of section') from e
            except configparser.ParsingError as e:
                lineno, msg = e.errors[0]
                # TODO: Remove this when Python 3.12 is deprecated.
                import sys
                if sys.version_info >= (3, 13):
                    msg = repr(msg)
                raise errors.ConfigError(f'{filepath}: Line {lineno}: {msg}: Invalid syntax') from e
            except configparser.DuplicateSectionError as e:
                raise errors.ConfigError(f'{filepath}: Line {e.lineno}: {e.section}: Duplicate section') from e
            except configparser.DuplicateOptionError as e:
                raise errors.ConfigError(f'{filepath}: Line {e.lineno}: {e.option}: Duplicate option') from e
            except configparser.Error as e:
                raise errors.ConfigError(f'{filepath}: {e}') from e

        # Convert ConfigParser instance to normal `dict`.
        # https://stackoverflow.com/a/28990982
        cfg = {
            section_name: dict(cfgparser.items(section_name))
            for section_name in cfgparser.sections()
        }

        # Make sure names can be valid object attributes, i.e. rename "foo-bar" to "foo_bar".
        def fix_name(name):
            return name.replace('-', '_')

        for section_name in tuple(cfg):
            # Fix section names.
            section_name_fixed = fix_name(section_name)
            cfg[section_name_fixed] = cfg.pop(section_name)
            # Fix option names in section.
            section = cfg[section_name_fixed]
            for option_name in tuple(section):
                option_name_fixed = fix_name(option_name)
                section[option_name_fixed] = section.pop(option_name)

        # Line breaks are interpreted as list separators
        for section in cfg.values():
            for option in section:
                if '\n' in section[option]:
                    section[option] = tuple(
                        item
                        for item in section[option].split('\n')
                        if item
                    )

        return cfg

    def _read_file(self, filepath, *, ignore_missing=False):
        try:
            with open(filepath, 'r') as f:
                return f.read()
        except OSError as e:
            if isinstance(e, FileNotFoundError) and ignore_missing:
                return ''
            else:
                msg = e.strerror if e.strerror else str(e)
                raise errors.ConfigError(f'{filepath}: {msg}') from e

    def write(self, *sections, include_defaults=False):
        """
        Save current configuration to :attr:`files`

        List values use ``"\\n "`` (newline followed by two spaces) as separators between items.

        :param sections: Sections to write (see :attr:`files`)
            Write all sections if no sections are provided.

        :param bool include_defaults: Whether to include commented defaults

        :raise ConfigError: if writing fails
        """
        if not sections:
            sections = tuple(self.files)

        for section in sections:
            # Generate file content.
            ini_string = (
                self[section].as_ini(include_defaults=include_defaults)
                + '\n'
            )

            # Make sure parent directory exists with the correct permissions.
            parentdir = os.path.dirname(self.files[section])
            if parentdir:
                os.makedirs(parentdir, mode=0o700, exist_ok=True)
            try:
                with open(self.files[section], 'w') as f:
                    f.write(ini_string)
            except OSError as e:
                msg = e.strerror if e.strerror else str(e)
                raise errors.ConfigError(f'{self.files[section]}: {msg}') from e


class SectionBase(ConfigDictBase):
    """
    Configuration model of one section

    Instances map subsection names to :class:`SubsectionBase` instances.

    From a user perspective, a subsection a section in an INI file, e.g. "[qbittorrent]" in
    ~/.config/upsies/clients.ini.
    """

    def __getitem__(self, key):
        try:
            return super().__getitem__(key)
        except KeyError:
            raise errors.UnknownSubsectionConfigError(key) from None

    def validate_config(self, config):
        """
        Called by :meth:`ConfigBase.validate_config` with its own instance after validation

        This method does nothing by default. Subclasses can override it if access to sibling
        sections is required for validation.
        """

    def as_ini(self, *, include_defaults=False):
        """Return section in INI file format"""
        parts = []
        for subsection in self:
            parts.extend(self._as_ini(subsection, include_defaults=include_defaults))
        return '\n'.join(parts).strip()

    def _as_ini(self, subsection, *, include_defaults):
        options = self[subsection].as_ini(include_defaults=include_defaults)
        if options:
            yield f'[{subsection}]'
            yield options
            yield ''  # Empty line as separator between sections.


class SubsectionBase(ConfigDictBase):
    """
    Configuration model of one subsection

    Instances map option names to values.

    From a user perspective, an option is a single key-value pair in a section in an INI file,
    e.g. "foo = bar".
    """

    def __getitem__(self, key):
        try:
            return super().__getitem__(key)
        except KeyError:
            raise errors.UnknownOptionConfigError(key) from None

    def get_description(self, option):
        """Return the description for `option` or empty string"""
        description = type(self).__pydantic_fields__[option].description
        if description:
            return string.evaluate_fstring(
                description,
                config=self,
                __project_name__=__project_name__,
            )
        else:
            return ''

    def as_ini(self, *, include_defaults=False, comment_defaults=True):
        """Return option-value pairs in INI file format"""
        return '\n'.join(self._as_ini(
            include_defaults=include_defaults,
            comment_defaults=comment_defaults,
        ))

    def _as_ini(self, *, include_defaults=False, comment_defaults=True):
        for option in self:
            yield from self._as_ini_option_lines(
                option,
                include_default=include_defaults,
                comment_default=comment_defaults,
            )

    def _as_ini_option_lines(self, option, *, include_default, comment_default):
        value = self[option]

        def option_lines(*lines):
            yield from self._as_ini_option_lines_filter(
                option=option,
                lines=lines,
                include_default=include_default,
                comment_default=comment_default,
            )

        if utils.is_sequence(value):
            # One item per line with 2 spaces indentation.
            yield from option_lines(
                f'{option} =',
                *(f'  {item}' for item in value),
            )

        elif isinstance(value, pydantic.SecretStr):
            # Values like password are not available unless explicitly requested to prevent leaks in
            # log files and such.
            yield from option_lines(f'{option} = {value.get_secret_value()}')

        else:
            yield from option_lines(f'{option} = {value}')

    def _as_ini_option_lines_filter(self, *, option, lines, include_default, comment_default):
        value = self[option]
        default = self.defaults[option]
        if value != default or not comment_default:
            # Custom value.
            yield from lines

        elif include_default:
            # Commented default value.
            for line in lines:
                yield '# ' + line



"""
Manage configuration files
"""

from .. import errors, utils
from . import JobBase

import logging  # isort:skip
_log = logging.getLogger(__name__)


class SetJob(JobBase):
    """Display or change configuration file option"""

    name = 'set'
    label = 'Set'
    hidden = True

    # Don't read output from cache.

    # TODO: Currently there is no way to prevent output from being written. We could do this by
    #       implementing an Enum with the values "NO_READ" and "NO_READ_WRITE".
    cache_id = None

    def initialize(self, *, config, option=None, value=(), reset=None, dump=()):
        """
        Set and display option(s)

        :param config: :class:`~.config.ConfigBase` instance

        :param str option: "."-delimited path to option in `config` or `None`

        :param value: New value for `option` or any falsy value to display the current value

            If `value` is a sequence and `option` is not, the items are joined with a space.
            If `value` is not a sequence and `option` is, `value` is turned into a sequence with the
            single item `value`.

        :param bool reset: Whether to reset `option` to default value and ignore `value`

        :param bool dump: Sequence of sections in `config`

            Read and write configuration the :attr:`~.ConfigBase.files` associated with each
            section. Default values are commented out so the user can easily edit each file.

        If only `config` is given, display all options and values.

        If `option` is given, only display its value.

        If `option` and `value` is given, set `option` to `value` and display the new value.

        If `dump` is given, it is a sequence of sections. For each section, the corresponding file
        is written.
        """
        if option and dump:
            raise RuntimeError('Arguments "option" and "dump" are mutually exclusive.')

        if value:
            if reset:
                raise RuntimeError('Arguments "value" and "reset" are mutually exclusive.')
            if dump:
                raise RuntimeError('Arguments "value" and "dump" are mutually exclusive.')

        self._config = config
        self._option = option
        self._value = value
        self._reset = reset
        self._dump = dump

    async def run(self):
        try:
            if self._reset:
                self._reset_mode()
            elif self._value:
                self._set_mode()
            elif self._dump:
                self._dump_mode()
            else:
                self._display_mode()

        except errors.UnknownConfigError as e:
            # UnknownConfigError is a subclass of KeyError, so ``str(e)`` is quoted and so the
            # actual error message has to come from an attribute.
            self.error(e.message)

        except errors.ConfigError as e:
            self.error(e)

    def _display_mode(self):
        self._display_option(self._option)

    def _set_mode(self):
        self._set_option(self._option, self._value)
        self._write_section(self._option)
        self._display_option(self._option)

    def _reset_mode(self):
        path = self._normalize_path(self._option)
        section, subsection, option, *_ = (*path, None, None, None)

        try:
            if section and subsection and option:
                _log.debug('Resetting option: %s', '.'.join((section, subsection, option)))
                default = self._config[section][subsection].defaults[option]
                self._config[section][subsection][option] = default
            elif section and subsection:
                _log.debug('Resetting subsection: %s', '.'.join((section, subsection)))
                default = self._config[section][subsection].defaults
                self._config[section][subsection] = default
            elif section:
                _log.debug('Resetting section: %s', section)
                default = self._config[section].defaults
                self._config[section] = default
            else:
                _log.debug('Resetting all sections')
                self._config = self._config.defaults
        except errors.UnknownConfigError as e:
            # Prepend the path to the error message.
            raise errors.ConfigError('.'.join(path) + f': {e.message}') from e
        else:
            self._write_section(self._option)
            self._display_option(self._option)

    def _dump_mode(self):
        for section in self._dump:
            # Write file
            self._config.write(section, include_defaults=True)

            # Display file
            banner_width = 78
            self.add_output('#' * banner_width)
            msg = f' Wrote {self._config.files[section]} '
            self.add_output('###' + msg.ljust(banner_width - 3, '#'))
            self.add_output('#' * banner_width)
            with open(self._config.files[section], 'r') as f:
                self.add_output(f.read())

    def _display_option(self, path):
        """
        Show value(s) of option, subsection or section referred to by `path`

        :param path: See :meth:`_get_valid_path_and_value`.
        """
        def display(value, path):
            if isinstance(value, utils.config.ConfigDictBase):
                # Display section or subsection of options.
                for option_name, value_ in value.items():
                    display(value_, path=(*path, option_name))

            elif utils.is_sequence(value):
                value_string = '\n  '.join(str(v) for v in value)
                if value_string:
                    value_string = '\n  ' + value_string
                self.add_output('.'.join(path) + ' =' + value_string)
            else:
                value_string = ' ' + str(value)
                self.add_output('.'.join(path) + ' =' + value_string)

        parent_path, value = self._get_valid_path_and_value(path)
        display(value, parent_path)

    def _set_option(self, path, value):
        """
        Set option referred to by `path` to `value`

        :param path: See :meth:`_get_valid_option_path_and_value`.
        """
        (section, subsection, option), current_value = self._get_valid_option_path_and_value(path)

        # Make sure `value` is (not) a sequence, depending on current value. Conversion from `str`
        # should be automatically handled by `self._config`.
        if utils.is_sequence(current_value) and not utils.is_sequence(value):
            value = (value,)
        elif not utils.is_sequence(current_value) and utils.is_sequence(value):
            value = ' '.join(value)

        # Set new value.
        self._config[section][subsection][option] = value

        # Some options need the whole config for validation, e.g. if `trackers.*.add_to` is set to
        # "foo", `TrackersConfig` must check if `clients.foo` exists.
        self._config.validate_config()

        return '.'.join(path), self._config[section][subsection][option]

    def _write_section(self, path):
        """
        Write section to its associated file (see :meth:`~.ConfigBase.write`)

        :param path: See :meth:`_normalize_path`.
        """
        section_name, *_ = self._normalize_path(path)
        self._config.write(section_name)

    def _get_valid_path_and_value(self, path):
        """
        Return normalized and validated option/subsection/section path and the corresponding
        value/subsection/section

        :param path: See :meth:`_normalize_path`.

        :raise ConfigError: if `path` does not refer to an existing option, subsection or section
        """
        path = self._normalize_path(path)
        value = self._config
        valid_path = []
        for key in path:
            try:
                value = value[key]
            except errors.UnknownConfigError as e:
                raise errors.ConfigError('.'.join(path) + f': {e.message}') from e
            else:
                valid_path.append(key)
        return (tuple(valid_path), value)

    def _get_valid_option_path_and_value(self, path):
        """
        Return normalized and validated option path and corresponding value

        :param path: See :meth:`_normalize_path`.

        :raise ConfigError: if `path` does not refer to an existing option
        """
        path, value = self._get_valid_path_and_value(path)
        if len(path) == 0:
            raise errors.ConfigError('Missing section, subsection and option')
        elif len(path) == 1:
            raise errors.ConfigError(f'Missing subsection and option in section {path[0]}')
        elif len(path) == 2:
            raise errors.ConfigError(f'Missing option in subsection {path[0]}.{path[1]}')
        else:
            return path, value

    def _normalize_path(self, path):
        """
        Return any valid path, e.g. ``(section,)``, ``(section, subsection)`` or ``(section,
        subsection, option)``

        :param path: :class:`str` (``section.subsection.option``, ``section.subsection`` or
            ``section``) or any valid return value of this method

        .. note:: This method does only return ``path`` in a valid format. It does not check if
            ``path`` resolves to an existing option, subsection or section.

        :raise ConfigError: if `path` is not a valid path to a section, subsection or option
        """
        if not path:
            return ()
        else:
            if not isinstance(path, (tuple, list)):
                path = tuple(k for k in path.split('.') if k.strip())
            if len(path) > 3:
                raise errors.ConfigError('.'.join(path) + ': Invalid option')
            else:
                return path



"""
API for trackers

:class:`~.TrackerBase` provides a uniform interface for all trackers. Its subclasses have two main
purposes:

1. Specify jobs that generate metadata, e.g. torrent creation, ``mediainfo`` output, IMDb ID, etc.

2. Provide coroutine methods, e.g. for uploading the generated metadata.
"""

import pydantic

from .. import errors, trackers, utils
from .base import TrackerBase, TrackerConfigBase, TrackerJobsBase, rules


def tracker_classes():
    """Return list of :class:`.TrackerBase` subclasses"""
    return utils.subclasses(TrackerBase, utils.submodules(__package__))


def tracker(name, **kwargs):
    """
    Create :class:`.TrackerBase` instance

    :param str name: Name of the tracker. A subclass of :class:`.TrackerBase`
        with the same :attr:`~.TrackerBase.name` must exist in one of this
        package's submodules.
    :param kwargs: All keyword arguments are passed to the subclass specified by
        `name`

    :raise ValueError: if no matching subclass can be found

    :return: :class:`.TrackerBase` instance
    """
    for cls in tracker_classes():
        if cls.name == name:
            return cls(**kwargs)
    raise ValueError(f'Unsupported tracker: {name}')


def tracker_names():
    """Return sequence of valid `name` arguments for :func:`.tracker`"""
    return sorted(utils.CaseInsensitiveString(cls.name) for cls in tracker_classes())


def _trackers_config_fields():
    """
    Return :class:`dict` that maps field names (:attr:`.TrackerBase.name`) to
    `(<TrackerConfigBase subclass>, <TrackerConfigBase instance>)` tuples
    """
    clses = sorted(
        tracker_classes(),
        key=lambda cls: cls.name,
    )
    return {
        cls.name: (cls.TrackerConfig, cls.TrackerConfig())
        for cls in clses
    }


class _TrackersConfig(utils.config.SectionBase):
    def validate_config(self, config):
        known_clients = set(config['clients'])
        known_clients.update(config['clients'].defaults)
        for tracker_name, tracker_config in self.items():
            add_to = tracker_config['add_to']
            if add_to and add_to not in known_clients:
                raise errors.ConfigValueError(f'No such client: {add_to}', path=('trackers', tracker_name, 'add_to'))


TrackersConfig = pydantic.create_model(
    'TrackersConfig',
    **_trackers_config_fields(),
    __base__=_TrackersConfig,
)



import abc
import collections
import typing

import aiobtclientapi
import pydantic

from .. import errors, utils
from . import fields

import logging  # isort:skip
_log = logging.getLogger(__name__)


class _GetConfigAttribute:
    """Get ``name`` and ``label`` from :attr:`~.Btclient.Config`"""

    def __init__(self, attribute_name):
        self.attribute_name = attribute_name

    def __get__(self, obj, objtype):
        return getattr(objtype.Config, self.attribute_name)


class Btclient:
    """
    Base class for BitTorrent client APIs

    This is a wrapper around a :class:`aiobtclientapi.APIBase` subclass.

    :param config: :class:`~.BtclientConfig` instance
    """

    @abc.abstractmethod
    def Config(self):
        """:class:`~.BtclientConfig` subclass that must be set by subclasses"""

    @classmethod
    def from_config(cls, config):
        """
        Return subclass instance where :attr:`name` matches ``config["client"]``

        :raise ConfigError: if no match is found
        """
        # Find subclass by matching "client" against each subclass' "name" attribute.
        config_client = config['client']
        for subcls in Btclient.__subclasses__():
            if config_client == subcls.name:
                return subcls(config)

        raise errors.ConfigError('No client specified')

    name = _GetConfigAttribute('name')
    """
    Name of the client (in configuration files and internal use)

    See :attr:`~.BtclientConfig.client`.
    """

    label = _GetConfigAttribute('label')
    """
    Official name of the client (e.g. in documentation)

    See :attr:`~.BtclientConfig.label`.
    """

    @classmethod
    def client_names(cls):
        """Sequence of :attr:`name` values of subclasses"""
        return fields.KNOWN_CLIENT_NAMES

    def __init__(self, config):
        assert isinstance(config, self.Config), f'Not a {self.Config.__name__}: {config!r}'
        self._config = config
        self._api = aiobtclientapi.api(
            name=config['client'],
            url=config['url'],
            username=config['username'],
            password=config['password'].get_secret_value(),
        )

    def __repr__(self):
        return f'{type(self).__name__}({self._config!r})'

    async def add_torrent(self, torrent, *, download_path=None):
        """
        Add `torrent` to client

        :param torrent: ``.torrent`` file/URL, magnet link or infohash
        :param download_path: Where the files in `torrent` can be found or should be downloaded to
        """
        _log.debug('Adding %s to %s with config=%s', torrent, self.name, self._config)
        response = await self._api.add(
            torrent,
            location=download_path,
            # Note that "check_after_add" option is not supported by all clients.
            verify=self._config.get('check_after_add', False),
        )

        # Raise first error.
        for error in response.errors:
            raise errors.TorrentAddError(error)

        # Return first infohash of first added (or already added) torrent.
        added = (response.added + response.already_added)
        if added:
            return added[0]

        raise RuntimeError(f'Failed to add torrent: {torrent}')


class BtclientConfig(utils.config.SubsectionBase):
    """
    Bittorrent client configuration base class
    """

    # Immutable class attribute that identifies the BitTorrent client (e.g. "qbittorrent", "deluge",
    # etc). Must be specified by subclass.
    name: fields.name

    # Same as `name`, but mutable and stored on the instance. This is used to derive a subclass.
    client: fields.client

    # Proper name of the BitTorrent client (e.g. "qBittorrent", "Deluge", etc).
    # Must be specified by subclass.
    label: fields.label

    def __new__(cls, **kwargs):
        # Get the actual class we want to instantiate.
        guessed_cls = cls.guess_subclass(kwargs)
        self = super().__new__(guessed_cls)

        if not issubclass(guessed_cls, cls):
            # __init__() is only called automatically if we return an instance of `cls`, which is
            # not the case if, for example, we guessed `QbittorrentBtclientConfig` from
            # `GenericBtclientConfig`.
            self.__init__(**kwargs)

        return self

    @classmethod
    def guess_subclass(cls, kwargs):
        """Return correct subclass based on "client" option in `kwargs` or :attr:`name`"""
        # Do not guess the correct subclass if this is a `*Defaults` class (see
        # `ConfigDictBase.defaults`).
        if cls.__name__.endswith('Defaults'):
            return cls

        def get_client_name(subcls, kws=None):
            # Get client name from keyword arguments. Keyword arguments always take precedence over
            # the class attribute so we can do `DelugeBtclientConfig(client="transmission")` to get
            # a `TransmissionBtclientConfig` instance. This is ugly, but it's necessary to allow the
            # user to change the "client" option without editing the config file.
            client_name = None
            if kws:
                client_name = kws.get('client', None)

            if not client_name:
                # Get client name from class attribute.
                client_name = getattr(subcls, 'name', None)

            return client_name

        wanted_client_name = get_client_name(cls, kwargs)
        subclses = BtclientConfig.__subclasses__()
        for subcls in subclses:
            subcls_client_name = get_client_name(subcls)
            if wanted_client_name == subcls_client_name:
                return subcls

        return GenericBtclientConfig

    def convert(self):
        """
        Return different :class:`BtclientConfig` instance based on "client" field

        Return `None` if not proper subclass can be determined (i.e. the determined subclass is
        :class:`GenericBtclientConfig`) or if the "client" field has its default value.

        For example:

        >>> c1 = DelugeBtclientConfig()
        >>> isinstance(c1, DelugeBtclientConfig)
        True
        >>> c1["client"] = 'transmission'
        >>> c2 = c1.convert()
        >>> isinstance(c2, TransmissionBtclientConfig)
        True
        >>> c1["client"] = 'deluge'
        >>> c1.convert() is None
        True
        """
        subclass = self.guess_subclass(dict(self))
        if (
                subclass is not GenericBtclientConfig
                and self['client'] != self.defaults['client']
        ):
            known_fields = set(subclass.__pydantic_fields__)
            kwargs = {
                key: value
                for key, value in self.items()
                if (
                        # Subclass knows what to do with `key`, e.g. "category" is only implemented
                        # for QbittorrentBtclientConfig, so we ignore it when converting to another
                        # client.
                        key in known_fields
                        # Do not pass our default values to the new subclass. It may have its own
                        # default for `key`.
                        and self[key] != self.defaults[key]
                )
            }
            try:
                return subclass(**kwargs)
            except pydantic.ValidationError as e:
                raise self._value_error(e) from e

    def as_ini(self, *, client_id, include_defaults=False, comment_defaults=True):
        return '\n'.join(self._as_ini(
            client_id=client_id,
            include_defaults=include_defaults,
            comment_defaults=comment_defaults,
        ))

    def _as_ini(self, *, client_id, include_defaults=False, comment_defaults=True):
        # Always specify "client" option if we are stored under a custom client ID (i.e. not one of
        # "deluge", "qbittorrent", etc).
        if client_id != type(self).name:
            yield from self._as_ini_option_lines('client', include_default=True, comment_default=False)

        for option in self:
            # Exclude "client" because we either included it above or we don't want it.
            if option != 'client':
                yield from self._as_ini_option_lines(
                    option,
                    include_default=include_defaults,
                    comment_default=comment_defaults,
                )


class GenericBtclientConfig(BtclientConfig):
    """
    Dummy :class:`BtclientConfig` that can be converted to a proper implementation

    Converting to a propery implementation is done by setting the "client" field and calling
    :meth:`~.BtclientConfig.convert`.
    """

    name: fields.name = ''
    client: fields.client = ''
    label: fields.label = ''


def _validate_client_config(cfg):
    if isinstance(cfg, BtclientConfig):
        return cfg
    elif isinstance(cfg, collections.abc.Mapping):
        # This will get us a `DelugeBtclientConfig` if `cfg["client"] == "deluge"` and default to
        # `GenericBtclientConfig`.
        return BtclientConfig(**cfg)
    else:
        raise TypeError(f'Not a {BtclientConfig.__name__} or Mapping: {type(cfg).__name__}: {cfg!r}')


class BtclientsConfig(utils.config.SectionBase, extra='allow'):
    """
    Map client IDs (:class:`str`) to :class:`~.BtclientConfig` instances

    From a user perspective, a client ID is a section name in "clients.ini", e.g. "[qbittorrent]",
    "[nas]" or "[seedbox]".

    Client IDs are arbitrary strings specified by the user. If a client ID is a valid
    :attr:`.Btclient.name`, it is used to derive the proper subclass from that, e.g. a section named
    "[qbitorrent]" in the INI file will result in a :class:`QbittorrentBtclientConfig` instance
    because :class:`QbittorrentBtclientConfig.name` is "qbittorrent".
    """

    __pydantic_extra__: dict[
        str,
        typing.Annotated[
            BtclientConfig,
            pydantic.BeforeValidator(_validate_client_config),
        ]
    ]

    def __init__(self, **kwargs):
        # If no "client" is specified, use the section name (e.g. "[deluge]" -> client="deluge").
        # The "client" field is used to derive a subclass, e.g. `DelugeBtclientConfig`.
        for section_name, section in kwargs.items():
            if (
                    'client' not in section
                    and section_name in Btclient.client_names()
            ):
                section['client'] = section_name
        super().__init__(**kwargs)

    def __getitem__(self, key):
        try:
            cfg = super().__getitem__(key)

        except errors.UnknownSubsectionConfigError:
            # Client IDs are arbitrary strings, so we can create them on demand.
            if key in Btclient.client_names():
                cfg = BtclientConfig(client=key)
            else:
                cfg = BtclientConfig()
            self[key] = cfg
            return cfg

        else:
            # If "client" is different from its default value, we have to convert, e.g. from
            # DelugeBtclientConfig to TransmissionBtclientConfig.
            if cfg['client'] != cfg.defaults['client']:
                try:
                    cfg_converted = cfg.convert()
                except errors.ConfigValueError as e:
                    # Conversion failed, e.g. because an option that is valid for
                    # TransmissionBtclientConfig is not valid for DelugeBtclientConfig.
                    raise errors.ConfigValueError(e.message, path=(key, *e.path)) from None
                else:
                    # If `cfg_converted` is `None`, conversion is not possible, e.g. because
                    # "client" option is not provided.
                    if cfg_converted is not None:
                        self[key] = cfg = cfg_converted
                    else:
                        # This should never happen because "client" cannot have any value that isn't
                        # a valid `Btname`. But convert() can return `None` in theory, so we raise
                        # RuntimeError for easier debugging if that happens.
                        raise RuntimeError(f'Not enough information to convert: {cfg!r}')

            return cfg

    @classmethod
    def get_defaults(cls):
        """
        Yield 2-tuples of `client_id` (:class:`str`) and :class:`~.BtclientConfig` subclasses

        We want the default instance of this class to be empty (no clients specified), but we also
        want :attr:`~.ConfigDictBase.defaults` to provide all available implementations (e.g. for
        documentation). This method gives us the latter. See also :class:`~._CreateDefaults`.
        """
        for subcls in Btclient.__subclasses__():
            yield subcls.Config.name, subcls.Config

    def as_ini(self, *, include_defaults=False):
        """Return section in INI file format"""
        # Custom client IDs.
        subsections = tuple(self)
        # Default client IDs.
        subsections += tuple(
            subsection
            for subsection in self.defaults
            if subsection not in subsections
        )
        parts = []
        for subsection in subsections:
            parts.extend(self._as_ini(subsection, include_defaults=include_defaults))
        return '\n'.join(parts).strip()

    def _as_ini(self, subsection, *, include_defaults):
        lines = self[subsection].as_ini(client_id=subsection, include_defaults=include_defaults).splitlines()
        if lines:
            if all(line.strip().startswith('#') for line in lines):
                yield f'# [{subsection}]'
            else:
                yield f'[{subsection}]'
            yield from lines
            yield ''



from .. import constants, utils


class MainConfig(utils.config.SubsectionBase):
    cache_directory: utils.config.fields.string(
        default=constants.DEFAULT_CACHE_DIRECTORY,
        description=(
            'Where to store generated files.\n'
            'WARNING: Files in this directory are automatically deleted until the '
            'whole directory is smaller than <max_cache_size> bytes.'
        ),
    )

    max_cache_size: utils.config.fields.custom(
        cls=utils.types.Bytes,
        default='100 MB',
        description=(
            'Maximum size of cache directory. '
            'Units like "kB" and "MiB" are interpreted.'
        ),
    )

    check_for_prerelease: utils.config.fields.boolean(
        default='no',
        description=(
            'Whether to notify you of new prereleases. '
            '(The version number of prereleases ends with "alpha".) '
            'If you are already running a prerelease, you will always '
            'get notified of new prereleases, and this option has no effect.'
        ),
    )


class IdConfig(utils.config.SubsectionBase):
    show_poster: utils.config.fields.boolean(
        default='no',
        description=(
            'Whether to display a poster for easier identification.'
        ),
    )


class ScreenshotsConfig(utils.config.SubsectionBase):
    optimize: utils.config.fields.choice(
        default='default',
        options=utils.image.optimization_levels,
        case_sensitive=False,
        description=(
            'How much CPU resources to put into screenshot size optimization.\n'
            'Valid values are: ' + ', '.join(utils.image.optimization_levels)
        ),
    )


class TorrentCreateConfig(utils.config.SubsectionBase):
    reuse_torrent_paths: utils.config.fields.custom(
        cls=utils.types.ListOf(str),
        default=(),
        description=(
            'List of directories to search for a *.torrent file to get piece hashes '
            'from when creating a torrent.'
        ),
    )


class ConfigConfig(utils.config.SectionBase):
    main: MainConfig = MainConfig()
    torrent_create: TorrentCreateConfig = TorrentCreateConfig()
    screenshots: ScreenshotsConfig = ScreenshotsConfig()
    id: IdConfig = IdConfig()



"""
Concrete :class:`~.TrackerConfigBase` subclass for UHD
"""

import base64

from ... import utils
from .. import base

UhdImagehost = utils.types.Imagehost(
    disallowed=(
        ()
        if utils.is_running_in_development_environment() else
        ('dummy',)
    ),
)


class UhdTrackerConfig(base.TrackerConfigBase):
    base_url: base.config.base_url(
        base64.b64decode('aHR0cHM6Ly91aGRiaXRzLm9yZw==').decode('ascii'),
    )

    username: base.config.username('')

    password: base.config.password('')

    cookies_filepath: base.config.cookies_filepath('')

    announce_url: base.config.announce_url('', autofetched=True)

    image_host: base.config.image_host(
        UhdImagehost,
        default=('ptpimg', 'freeimage', 'imgbox'),
    )

    screenshots_count: base.config.screenshots_count(4, min=2, max=10)

    exclude: base.config.exclude(
        base.exclude_regexes.checksums,
        base.exclude_regexes.images,
        base.exclude_regexes.nfo,
        base.exclude_regexes.samples,
        base.exclude_regexes.subtitles,
    )

    anonymous: base.config.anonymous('no')


cli_arguments = {
    'submit': {
        ('--imdb', '--im'): {
            'help': 'IMDb ID or URL',
            'type': utils.argtypes.webdb_id('imdb'),
        },
        ('--anonymous', '--an'): {
            'help': 'Hide your username for this submission',
            'action': 'store_true',
            # This must be `None` so it doesn't override the "anonymous"
            # value from the config file. See CommandBase.get_options().
            'default': None,
        },
        ('--internal', '--in'): {
            'help': 'Internal encode (use only if you were told to)',
            'action': 'store_true',
        },
        ('--3d',): {
            'help': 'Mark this as a 3D release',
            'action': 'store_true',
        },
        ('--vie', '--vi'): {
            'help': 'Release contains Vietnamese audio dub',
            'action': 'store_true',
        },
        ('--screenshots-count', '--ssc'): {
            'help': ('How many screenshots to make '
                     f'(min={UhdTrackerConfig.defaults["screenshots_count"].min}, '
                     f'max={UhdTrackerConfig.defaults["screenshots_count"].max})'),
            'type': utils.argtypes.number_of_screenshots(
                min=UhdTrackerConfig.defaults['screenshots_count'].min,
                max=UhdTrackerConfig.defaults['screenshots_count'].max,
            ),
        },
        ('--screenshots', '--ss'): {
            'help': (
                'Path(s) to existing screenshot file(s)\n'
                'Directories are searched recursively.\n'
                'More screenshots are created if necessary.'
            ),
            'nargs': '+',
            'action': 'extend',
            'type': utils.argtypes.files_with_extension('png'),
            'metavar': 'SCREENSHOT',
        },
        ('--nfo',): {
            'help': 'Path to NFO file (supersedes any *.nfo file found in the release directory)',
        },
        ('--poster', '--po'): {
            'help': 'Path or URL to poster image (autodetected by default)',
        },
        ('--trailer', '--tr'): {
            'help': 'Trailer YouTube ID or URL (autodetected by default)',
        },
        ('--only-description', '--od'): {
            'help': 'Only generate description (do not submit)',
            'action': 'store_true',
            'group': 'generate-metadata',
        },
    },
}



"""
Base class for tracker configuration
"""

from ... import btclients, constants, utils

import logging  # isort:skip
_log = logging.getLogger(__name__)


def base_url(default):
    return utils.config.fields.string(
        default=default,
        description='Base URL of the website.\n(Changing this is only useful for debugging purposes.)',
    )


def upload_url(default):
    return utils.config.fields.string(
        default=default,
        description='Upload API URL.\n(Changing this is only useful for debugging purposes.)',
    )


def announce_url(default, *, autofetched=False, instructions=''):
    description = 'Your personal announce URL.'
    if autofetched:
        description += '\nAutomatically fetched from the website if not set.'
    if instructions:
        description += '\n' + str(instructions)
    return utils.config.fields.string(
        default=default,
        description=description,
    )


def apikey(default, *, instructions=''):
    description = 'Your personal API key.'
    if instructions:
        description += '\n' + str(instructions)
    return utils.config.fields.string(
        default=default,
        description=description,
    )


def username(default):
    return utils.config.fields.string(
        default=default,
        description='Your personal username.',
    )


def password(default):
    return utils.config.fields.string(
        default=default,
        description='Your personal password.',
        secret=True,
    )


def exclude(*defaults):
    return utils.config.fields.custom(
        cls=utils.types.ListOf(utils.types.Regex),
        default=defaults,
        description='List of regular expressions. Matching files are excluded from generated torrents.',
    )


def anonymous(default):
    return utils.config.fields.boolean(
        default=default,
        description='Whether your username is displayed on your uploads.',
    )


def image_host(Imagehost, *, default):
    return utils.config.fields.custom(
        cls=utils.types.ListOf(Imagehost, separator=','),
        default=default,
        description=(
            'List of image hosting service names. The first service is normally used '
            'with the others as backup if uploading fails.\n'
            'Supported services: ' + ', '.join(Imagehost.options)
        ),
    )


def screenshots_count(default, *, min, max, description=''):
    if not description:
        description = 'How many screenshots to make.'
    description += f'\nMinimum is {min}, maximum is {max}.'
    return utils.config.fields.integer(
        default=default,
        min=min,
        max=max,
        description=description,
    )


def cookies_filepath(default):
    return utils.config.fields.string(
        default=default,
        description=(
            'File that stores permanent session cookies.\n'
            'If this is not set, a new login session is started for each submission.'
        ),
    )


class TrackerConfigBase(utils.config.SubsectionBase):
    """
    Base class for a tracker's user configuration (e.g. via config file or CLI arguments)

    Options defined in this class are available for all trackers. Subclasses may add more options or
    override values of this class' options.
    """

    exclude: exclude()

    add_to: utils.config.fields.string(
        default='',
        description=(
            'BitTorrent client to add torrent to after successful submission.\n'
            'Valid values are ' + ", ".join(f'"{c}"' for c in btclients.client_names()) + ' or one of the section names '
            f'in {utils.fs.tildify_path(constants.CLIENTS_FILEPATH)}.'
        ),
    )

    copy_to: utils.config.fields.string(
        default='',
        description='Directory path to copy torrent file to after successful submission.',
    )



"""
Concrete :class:`~.TrackerConfigBase` subclass for MTV
"""

import base64

from ... import utils
from .. import base

MtvImagehost = utils.types.Imagehost(
    allowed=(
        ('imgbox', 'ptpimg', 'dummy')
        if utils.is_running_in_development_environment() else
        ('imgbox', 'ptpimg')
    ),
)


class MtvTrackerConfig(base.TrackerConfigBase):
    base_url: base.config.base_url(
        base64.b64decode('aHR0cHM6Ly93d3cubW9yZXRoYW50di5tZQ==').decode('ascii'),
    )

    username: base.config.username('')

    password: base.config.password('')

    cookies_filepath: base.config.cookies_filepath('')

    announce_url: base.config.announce_url('', autofetched=True)

    image_host: base.config.image_host(
        MtvImagehost,
        default=('imgbox',),
    )

    screenshots_count: base.config.screenshots_count(4, min=3, max=10)

    exclude: base.config.exclude(
        base.exclude_regexes.checksums,
        base.exclude_regexes.extras,
        base.exclude_regexes.images,
        base.exclude_regexes.nfo,
        base.exclude_regexes.samples,
        base.exclude_regexes.subtitles,
    )

    anonymous: base.config.anonymous('no')


cli_arguments = {
    'submit': {
        ('--imdb', '--im'): {
            'help': 'IMDb ID or URL',
            'type': utils.argtypes.webdb_id('imdb'),
        },
        ('--anonymous', '--an'): {
            'help': 'Hide your username for this submission',
            'action': 'store_true',
            # This must be `None` so it doesn't override the "anonymous"
            # value from the config file. See CommandBase.get_options().
            'default': None,
        },
        ('--screenshots-count', '--ssc'): {
            'help': ('How many screenshots to make '
                     f'(min={MtvTrackerConfig.defaults["screenshots_count"].min}, '
                     f'max={MtvTrackerConfig.defaults["screenshots_count"].max})'),
            'type': utils.argtypes.number_of_screenshots(
                min=MtvTrackerConfig.defaults['screenshots_count'].min,
                max=MtvTrackerConfig.defaults['screenshots_count'].max,
            ),
        },
        ('--screenshots', '--ss'): {
            'help': (
                'Path(s) to existing screenshot file(s)\n'
                'Directories are searched recursively.\n'
                'More screenshots are created if necessary.'
            ),
            'nargs': '+',
            'action': 'extend',
            'type': utils.argtypes.files_with_extension('png'),
            'metavar': 'SCREENSHOT',
        },
        ('--only-description', '--od'): {
            'help': 'Only generate description (do not submit)',
            'action': 'store_true',
            'group': 'generate-metadata',
        },
        ('--only-title', '--ot'): {
            'help': 'Only generate title (do not submit)',
            'action': 'store_true',
            'group': 'generate-metadata',
        },
        ('--ignore-dupes', '--id'): {
            'help': 'Force submission even if the tracker reports duplicates',
            'action': 'store_true',
        },
    },
}



"""
Image uploader for freeimage.host
"""

import json

from .. import errors, utils
from . import base

import logging  # isort:skip
_log = logging.getLogger(__name__)


class FreeimageImagehostConfig(base.ImagehostConfigBase):
    """User configuration for :class:`FreeimageImagehost`"""

    base_url: utils.config.fields.string(
        default='https://freeimage.host',
        description='Base URL of the API.',
    )
    apikey: utils.config.fields.string(
        secret=True,
        default='6d207e02198a847aa98d0a2a901485a5',
        description=(
            'API access key. '
            'The default value is the public API key from https://freeimage.host/page/api.'
        ),
    )


class FreeimageImagehost(base.ImagehostBase):
    """Upload images to freeimage.host"""

    name = 'freeimage'

    Config = FreeimageImagehostConfig

    async def _upload_image(self, image_path):
        try:
            response = await utils.http.post(
                url=self.config['base_url'] + '/api/1/upload',
                cache=False,
                data={
                    'key': self.config['apikey'],
                    'action': 'upload',
                    'format': 'json',
                },
                files={
                    'source': image_path,
                },
            )
        except errors.RequestError as e:
            # Error response is undocumented. I looks like this:
            # {
            #   "status_code": 400,
            #   "error": {
            #     "message": "Can't get target upload source info",
            #     "code": 310,
            #     "context": "CHV\\UploadException"
            #   },
            #   "status_txt": "Bad Request"
            # }
            try:
                info = json.loads(e.text)
                raise errors.RequestError(f'{info["status_txt"]}: {info["error"]["message"]}')
            except (TypeError, ValueError, KeyError):
                if e.text:
                    raise errors.RequestError(f'Upload failed: {e.text}') from e
                else:
                    raise errors.RequestError(f'Upload failed: {e}') from e

        _log.debug('%s: Response: %r', self.name, response)
        info = response.json()
        try:
            return info['image']['image']['url']
        except KeyError as e:
            raise RuntimeError(f'Unexpected response: {response}') from e



import typing

import aiobtclientapi
import pydantic

from .. import utils

KNOWN_CLIENT_NAMES = aiobtclientapi.client_names()


_ClientNameChoice = utils.types.Choice(
    options=KNOWN_CLIENT_NAMES,
    case_sensitive=False,
    empty_ok=True,
)


client = utils.config.fields.custom(
    cls=_ClientNameChoice,
    default='',
    description=(
        'BitTorrent client name\n'
        'Supported clients: ' + ', '.join(_ClientNameChoice.options)
    ),
)


name = typing.Annotated[
    typing.ClassVar[str],
    # `frozen=True` seems to have no effect. You can set `MyClass.name = "whatever"` without
    # exception. Not sure if this is a bug in pydantic or my thoughts.
    pydantic.Field(frozen=True),
]
"""Name of BitTorrent client that is used for comparison with :var:`client`"""


label = typing.Annotated[
    typing.ClassVar[str],
    pydantic.Field(frozen=True),
]
"""Official BitTorrent client name used in documentation, error messages, etc"""


def url(api):
    """
    Return :class:`~.typing.Annotated` ``api.url``

    :param api: :class:`aiobtclientapi.APIBase` subclass
    """
    return utils.config.fields.custom(
        cls=api.URL,
        default=api.URL.default,
        description=f'URL of the {api.label} RPC interface',
    )


username = utils.config.fields.string(
    default='',
    description='Username for authentication',
)


password = utils.config.fields.string(
    default='',
    secret=True,
    description='Password for authentication',
)


check_after_add = utils.config.fields.boolean(
    default='no',
    description='Whether added torrents should be hash checked',
)


translate_path = utils.config.fields.custom(
    cls=utils.types.PathTranslations,
    default=(
        # 'foo -> bar',
        # 'this -> that',
    ),
    description=(
        'Translate absolute paths on the computer that is running {__project_name__} (LOCAL) '
        'to paths on the computer that is running {config["label"]} (REMOTE)\n'
        'This is a list where LOCAL and REMOTE are separated by "->". Spaces are trimmed. '
        'When adding a torrent, LOCAL in the content path is replaced with REMOTE to get '
        "the path where the BitTorrent client can find the torrent's files.\n"
        'Example:\n'
        'clients.{config["client"]}.translate_path =\n'
        '  /home/me/My Projects -> /storage/seed_forever\n'
        '  /media/me/USB/ -> /storage/seed_temporarily\n'
    ),
)



"""
Concrete :class:`~.TrackerConfigBase` subclass for BHD
"""

import base64

from ... import utils
from .. import base

BhdImagehost = utils.types.Imagehost(
    allowed=(
        ('imgbox', 'ptpimg', 'imgbb', 'dummy')
        if utils.is_running_in_development_environment() else
        ('imgbox', 'ptpimg', 'imgbb')
    ),
)


class BhdTrackerConfig(base.TrackerConfigBase):
    upload_url: base.config.upload_url(
        base64.b64decode('aHR0cHM6Ly9iZXlvbmQtaGQubWUvYXBpL3VwbG9hZA==').decode('ascii')
    )

    announce_url: utils.config.fields.string(
        default=base64.b64decode('aHR0cHM6Ly9iZXlvbmQtaGQubWUvYW5ub3VuY2U=').decode('ascii'),
        description='Announce URL without the private passkey.',
    )

    announce_passkey: utils.config.fields.string(
        default='',
        description=(
            'The private part of the announce URL.\n'
            'Get it from the website: My Security -> Passkey'
        )
    )

    apikey: base.config.apikey('')

    anonymous: base.config.anonymous('no')

    draft: utils.config.fields.boolean(
        default='no',
        description=(
            'Whether your uploads are stashed under Torrents -> Drafts '
            'after the upload instead of going live immediately.'
        ),
    )

    image_host: base.config.image_host(BhdImagehost, default=('imgbox',))

    screenshots_count: base.config.screenshots_count(4, min=3, max=10)

    exclude: base.config.exclude(
        base.exclude_regexes.checksums,
        base.exclude_regexes.extras,
        base.exclude_regexes.images,
        base.exclude_regexes.nfo,
        base.exclude_regexes.samples,
        base.exclude_regexes.subtitles,
    )


cli_arguments = {
    'submit': {
        ('--imdb', '--im'): {
            'help': 'IMDb ID or URL',
            'type': utils.argtypes.webdb_id('imdb'),
        },
        ('--tmdb', '--tm'): {
            'help': 'TMDb ID or URL',
            'type': utils.argtypes.webdb_id('tmdb'),
        },
        ('--anonymous', '--an'): {
            'help': 'Hide your username for this submission',
            'action': 'store_true',
            # This must be `None` so it doesn't override the "anonymous"
            # value from the config file. See CommandBase.get_options().
            'default': None,
        },
        ('--custom-edition', '--ce'): {
            'help': 'Non-standard edition, e.g. "Final Cut"',
            'default': '',
        },
        ('--draft', '--dr'): {
            'help': 'Upload as draft',
            'action': 'store_true',
            # The default value must be None so CommandBase.get_options()
            # doesn't always overwrite the value with the config file value.
            'default': None,
        },
        ('--nfo',): {
            'help': 'Path to NFO file (supersedes any *.nfo file found in the release directory)',
        },
        ('--personal-rip', '--pr'): {
            'help': 'Tag submission as your own encode',
            'action': 'store_true',
        },
        ('--screenshots-count', '--ssc'): {
            'help': ('How many screenshots to make '
                     f'(min={BhdTrackerConfig.defaults["screenshots_count"].min}, '
                     f'max={BhdTrackerConfig.defaults["screenshots_count"].max})'),
            'type': utils.argtypes.number_of_screenshots(
                min=BhdTrackerConfig.defaults['screenshots_count'].min,
                max=BhdTrackerConfig.defaults['screenshots_count'].max,
            ),
        },
        ('--screenshots', '--ss'): {
            'help': (
                'Path(s) to existing screenshot file(s)\n'
                'Directories are searched recursively.\n'
                'More screenshots are created if necessary.'
            ),
            'nargs': '+',
            'action': 'extend',
            'type': utils.argtypes.files_with_extension('png'),
            'metavar': 'SCREENSHOT',
        },
        ('--special', '--sp'): {
            'help': 'Tag as special episode, e.g. Christmas special (ignored for movie uploads)',
            'action': 'store_true',
        },
        ('--only-description', '--od'): {
            'help': 'Only generate description (do not submit)',
            'action': 'store_true',
            'group': 'generate-metadata',
        },
        ('--only-title', '--ot'): {
            'help': 'Only generate title (do not submit)',
            'action': 'store_true',
            'group': 'generate-metadata',
        },
    },
}



"""
Concrete :class:`~.TrackerConfigBase` subclass for NBL
"""

import base64

from ... import utils
from .. import base

import logging  # isort:skip
_log = logging.getLogger(__name__)


class NblTrackerConfig(base.TrackerConfigBase):
    upload_url: base.config.upload_url(
        base64.b64decode('aHR0cHM6Ly9uZWJ1bGFuY2UuaW8vdXBsb2FkLnBocA==').decode('ascii'),
    )

    apikey: base.config.apikey(
        default='',
        instructions='Create one on the website: <USERNAME> -> Settings -> API keys (check "Upload")',
    )

    announce_url: base.config.announce_url(
        default='',
        instructions='Get it from the website: Shows -> Upload -> Your personal announce URL',

    )

    exclude: base.config.exclude()

    anonymous: base.config.anonymous('no')


cli_arguments = {
    'submit': {
        ('--anonymous', '--an'): {
            'help': 'Hide your username for this submission',
            'action': 'store_true',
            # This must be `None` so it doesn't override the "anonymous"
            # value from the config file. See CommandBase.get_options().
            'default': None,
        },
        ('--tvmaze', '--tv'): {
            'help': 'TVmaze ID or URL',
            'type': utils.argtypes.webdb_id('tvmaze'),
        },
    },
}



"""
Dummy tracker for testing and debugging
"""

import asyncio
import functools
import os
import random
import string

from .. import errors, jobs, uis, utils
from . import base

import logging  # isort:skip
_log = logging.getLogger(__name__)


DummyImagehost = utils.types.Imagehost(allowed=('dummy',))


class DummyTrackerConfig(base.TrackerConfigBase):
    base_url: base.config.base_url('http://localhost')
    username: base.config.username('')
    password: base.config.password('')
    tfa: utils.config.fields.boolean(
        default='no',
        description='Whether to ask for 2FA one-time password for each login.',
    )

    image_host: base.config.image_host(
        DummyImagehost,
        default=('dummy',),
    )

    exclude: base.config.exclude()

    source: utils.config.fields.string(
        default='DMY',
        description='"source" field in generated torrents.',
    )


cli_arguments = {
    'submit': {
        ('--imdb', '--im'): {
            'help': 'IMDb ID or URL',
            'type': utils.argtypes.webdb_id('imdb'),
        },
        ('--tmdb', '--tm'): {
            'help': 'TMDb ID or URL',
            'type': utils.argtypes.webdb_id('tmdb'),
        },
        ('--skip-category', '-C'): {
            'help': 'Do not ask for category',
            'action': 'store_true',
        },
        ('--screenshots-count', '--ssc'): {
            'help': 'How many screenshots to make',
            'type': utils.argtypes.number_of_screenshots(min=3, max=10),
        },
        ('--screenshots', '--ss'): {
            'help': (
                'Path(s) to created screenshot file(s)\n'
                'Directories are searched recursively.\n'
                'More screenshots are created if necessary.'
            ),
            'nargs': '+',
            'action': 'extend',
            'type': utils.argtypes.files_with_extension('png'),
        },
        ('--delay', '-d'): {
            'help': 'Number of seconds login, upload and logout take each',
            'type': float,
            'default': 1.0,
        },
        ('--get-announce-from-website'): {
            'help': 'Whether to act like we are getting the announce URL from the website ',
            'action': 'store_true',
        },
        ('--poster'): {
            'help': 'Poster file or URL ',
        },
    },
    'torrent-create': {
        ('--delay', '-d'): {
            'help': 'Number of seconds login and logout take each',
            'type': float,
            'default': 0.0,
        },
        ('--get-announce-from-website'): {
            'help': 'Whether to act like we are getting the announce URL from the website ',
            'action': 'store_true',
        },
    },
}


class DummyTrackerJobs(base.TrackerJobsBase):

    release_name_separator = '_'

    @functools.cached_property
    def jobs_before_upload(self):
        return (
            self.login_job,

            # Interactive jobs
            self.playlists_job,
            self.tmdb_job,
            self.imdb_job,
            self.release_name_job,
            self.category_job,
            self.scene_check_job,

            # Background jobs
            self.create_torrent_job,
            self.screenshots_job,
            self.upload_screenshots_job,
            self.poster_job,
            self.mediainfo_job,

            # Silly jobs
            self.random_string_job,
            self.stupid_question_job,
            self.say_something_job,
            self.rules_job,
        )

    @functools.cached_property
    def random_string_job(self):
        return jobs.custom.CustomJob(
            name=self.get_job_name('random-string'),
            label='Random String',
            precondition=self.make_precondition('random_string_job'),
            worker=self.generate_random_string,
            catch=(
                errors.RequestError,
            ),
            **self.common_job_args(ignore_cache=True),
        )

    async def generate_random_string(self, job):
        # await asyncio.sleep(3)
        # self.random_string_job.warn('Watch out!')

        await asyncio.sleep(1)

        # raise errors.RequestError('foo :(')
        # raise RuntimeError('foo D:')

        return ''.join(
            random.choice(string.ascii_lowercase)
            for _ in range(30)
        )

    @functools.cached_property
    def stupid_question_job(self):
        return jobs.dialog.ChoiceJob(
            name=self.get_job_name('stupid-question'),
            label='Stupid Question',
            precondition=self.make_precondition('stupid_question_job'),
            prejobs=(
                self.random_string_job,
            ),
            autodetect=self.autodetect_stupid_answer,
            question='Is the random string random enough?',
            options=(
                'Possibly',
                'Maybe',
                'Perhaps',
            ),
            multichoice=True,
            validate=self._validate_chosen,
            callbacks={
                'finished': lambda job: job.clear_warnings(),
            },
            **self.common_job_args(),
        )

    def _validate_chosen(self, chosen):
        labels = [option[0] for option in chosen]
        if 'Perhaps' in labels:
            raise ValueError('I have decided that "Perhaps" is not an option after all.')
        elif len(labels) < 2:
            raise ValueError('Pick at least 2 options.')

    async def autodetect_stupid_answer(self, job):
        assert self.random_string_job.is_finished

        # raise RuntimeError('How would I know?')

        random_string = self.random_string_job.output[0]
        if all(character in random_string for character in 'abc'):
            return 'Maybe'
        else:
            self.stupid_question_job.add_prompt(
                uis.prompts.RadioListPrompt(
                    question='Is this job too stupid?',
                    options=('Yes', 'No'),
                    callbacks=(
                        self.prompt_stupid_answer_callback,
                    ),
                )
            )
            self.stupid_question_job.add_prompt(
                uis.prompts.CheckListPrompt(
                    question='Are you sure?',
                    options=('Yes', 'No', 'Maybe?'),
                    callbacks=(
                        self.prompt_stupid_answer_callback_2,
                    ),
                )
            )
            self.stupid_question_job.add_prompt(
                uis.prompts.RadioListPrompt(
                    options=('Foo', 'Bar'),
                    callbacks=(),
                )
            )

    def prompt_stupid_answer_callback(self, result):
        if result == 'Yes':
            self.stupid_question_job.error('This job is too stupid.')
        else:
            self.stupid_question_job.warn('Choose wisely!')

    def prompt_stupid_answer_callback_2(self, result):
        if 'Yes' in result and 'No' in result:
            self.stupid_question_job.error('Yes and No? Be more decisive!')
        elif 'Maybe?' in result:
            self.stupid_question_job.warn('Okay?')
        elif 'Yes' not in result:
            self.stupid_question_job.error('This job may be too stupid after all!')

    @functools.cached_property
    def say_something_job(self):
        return jobs.dialog.TextFieldJob(
            name=self.get_job_name('say-something'),
            label='Something',
            precondition=self.make_precondition('say_something_job'),
            prejobs=(
                self.random_string_job,
            ),
            warn_exceptions=(
                ValueError,
            ),
            text=self.generate_something,
            **self.common_job_args(),
        )

    async def generate_something(self):
        assert self.random_string_job.is_finished
        word = self.random_string_job.output[0]

        await asyncio.sleep(6)
        self.say_something_job.warn('Still thinking...')
        await asyncio.sleep(6)

        self.say_something_job.clear_warnings()
        if 'x' in word:
            raise ValueError(f'There is an "x" in "{word}"! Yuck!')
        elif 'a' in word:
            return 'Avocado!'
        else:
            return f'This is a completely original random string: {word}'

    @functools.cached_property
    def category_job(self):
        if not self.options['skip_category']:
            return jobs.dialog.ChoiceJob(
                name=self.get_job_name('category'),
                label='Category',
                precondition=self.make_precondition('category_job'),
                options=(
                    (str(typ).capitalize(), typ)
                    for typ in utils.types.ReleaseType if typ
                ),
                autodetected=self.release_name.type,
                **self.common_job_args(),
            )


class DummyHdOnly(base.rules.HdOnly):
    message = 'This is not HD!'

    async def _check(self):
        await super()._check()


class DummyBannedGroup(base.rules.BannedGroup):
    banned_groups = {
        'FoO',
        'BAR',
    }

    async def _check_custom(self):
        if (
                self.is_group('BAZ')
                and 'Remux' in self.release_name.source
        ):
            raise errors.BannedGroup('BAZ', additional_info='No remuxes')


class DummyNoEpisodes(base.rules.TrackerRuleBase):
    required_jobs = ('category_job', 'say_something_job', 'imdb_job')

    async def _check(self):
        _log.debug('################# CHECKING CATEGORY: %r', self.tracker_jobs.category_job.choice)
        await super()._check()
        if self.tracker_jobs.category_job.choice is utils.types.ReleaseType.episode:
            raise errors.RuleBroken('No episodes allowed')


class DummyTracker(base.TrackerBase):
    name = 'dummy'
    label = 'DuMmY'

    cli_arguments = cli_arguments
    rules = (
        DummyBannedGroup,
        DummyHdOnly,
        DummyNoEpisodes,
    )

    @property
    def torrent_source_field(self):
        return self.options['source']

    setup_howto_template = (
        'This is just a no-op tracker for testing and demonstration.'
    )

    TrackerJobs = DummyTrackerJobs
    TrackerConfig = DummyTrackerConfig

    async def _login(self, tfa_otp=None):
        username, password = self.options['username'], self.options['password']
        _log.debug('%s: Logging in as %s (password=%s, OTP=%s)', self.name, username, password, tfa_otp)
        if self.options['tfa'] and not tfa_otp:
            raise errors.TfaRequired('I need a a second factor.')

        await asyncio.sleep(self.options['delay'])

        # Simulate wrong OTP.
        if tfa_otp and not all(c in '1234567890' for c in tfa_otp):
            raise errors.RequestError('Your OTP looks weird.')

        # Simulate wrong username/password combination.
        if self.options['password'] == self.options['username']:
            raise errors.RequestError('Your username and password are identical, dummy!')

        _log.debug('%s: Logged in as %s (password=%s, OTP=%s)', self.name, username, password, tfa_otp)

    async def confirm_logged_in(self):
        _log.debug('%s: Confirming logged in', self.name)

    async def _logout(self):
        _log.debug('%s: Logging out', self.name)
        await asyncio.sleep(self.options['delay'])
        # raise errors.RequestError('You shall not logout!')
        _log.debug('%s: Logged out', self.name)

    async def get_announce_url(self):
        if self.options['get_announce_from_website']:
            _log.debug('%s: Getting announce URL from website', self.name)
            await asyncio.sleep(1)
        else:
            _log.debug('%s: Getting announce URL from config file', self.name)
        return 'http://localhost:123/f1dd15718/announce'

    async def upload(self, tracker_jobs):
        if tracker_jobs.create_torrent_job.output:
            torrent_file = tracker_jobs.create_torrent_job.output[0]
        else:
            raise errors.RequestError('Torrent file was not created.')
        _log.debug('%s: Uploading %s', self.name, torrent_file)
        await asyncio.sleep(self.options['delay'])
        return f'http://localhost/{os.path.basename(torrent_file)}'



import aiobtclientapi

from . import base, fields


class TransmissionBtclientConfig(base.BtclientConfig):
    name: fields.name = aiobtclientapi.TransmissionAPI.name
    label: fields.label = aiobtclientapi.TransmissionAPI.label
    client: fields.client = aiobtclientapi.TransmissionAPI.name

    url: fields.url(aiobtclientapi.TransmissionAPI)
    username: fields.username
    password: fields.password
    translate_path: fields.translate_path


class TransmissionBtclient(base.Btclient):
    Config = TransmissionBtclientConfig



"""
Concrete :class:`~.TrackerConfigBase` subclass for PTP
"""

import base64

from ... import utils
from .. import base
from . import metadata

PtpImagehost = utils.types.Imagehost(
    disallowed=(
        ()
        if utils.is_running_in_development_environment() else
        ('dummy',)
    ),
)


class PtpReleaseType(str):
    def __new__(cls, text):
        for type, regex in metadata.types.items():
            if regex.search(text):
                return super().__new__(cls, type)
        raise utils.argtypes.ArgumentTypeError(f'Invalid type: {text}')


class PtpTrackerConfig(base.TrackerConfigBase):
    base_url: base.config.base_url(
        base64.b64decode('aHR0cHM6Ly9wYXNzdGhlcG9wY29ybi5tZQ==').decode('ascii'),
    )

    username: base.config.username('')

    password: base.config.password('')

    cookies_filepath: base.config.cookies_filepath('')

    announce_url: base.config.announce_url('')

    image_host: base.config.image_host(
        PtpImagehost,
        default=('ptpimg', 'freeimage'),
    )

    screenshots_from_movie: base.config.screenshots_count(
        default=3,
        min=3,
        max=10,
        description='How many screenshots to make for single-video uploads.',
    )

    screenshots_from_episode: base.config.screenshots_count(
        default=2,
        min=2,
        max=10,
        description='How many screenshots to make per video for multi-video uploads.',
    )

    exclude: base.config.exclude(
        base.exclude_regexes.checksums,
        base.exclude_regexes.images,
        base.exclude_regexes.nfo,
        base.exclude_regexes.samples,
    )


cli_arguments = {
    'submit': {
        ('--subtitles', '--su'): {
            'help': (
                'Comma-separated list of subtitle language codes.\n'
                'Use this if subtitles are hardcoded or missing language tags.'
            ),
            'metavar': 'LANGUAGES',
            'type': utils.types.ListOf(
                item_type=utils.argtypes.subtitle,
                separator=',',
            ),
        },
        ('--hardcoded-subtitles', '--hs'): {
            'help': 'Release is trumpable because of hardcoded subtitles',
            'action': 'store_true',
        },
        ('--no-english-subtitles', '--nes'): {
            'help': (
                'Whether release contains no English audio and no English subtitles.\n'
                'This is autodetected reliably if all audio and subtitle tracks have '
                'a correct language tag. If not, you are asked interactively.\n'
                'Subtitle languages are detected in *.idx/sub, VIDEO_TS trees, '
                'BDMV trees and *.srt/ssa/ass/vtt by language code in the file name, e.g. '
                '"Foo.en.srt".'
            ),
            'metavar': 'BOOL',
            'type': utils.argtypes.bool_or_none,
        },
        ('--not-main-movie', '--nmm'): {
            'help': 'Upload ONLY contains extras, Rifftrax, Workprints',
            'action': 'store_true',
        },
        ('--personal-rip', '--pr'): {
            'help': 'Tag submission as your own encode',
            'action': 'store_true',
        },
        ('--poster', '--po'): {
            'help': 'Path or URL to movie poster',
        },
        ('--nfo',): {
            'help': 'Path to NFO file (supersedes any *.nfo file found in the release directory)',
        },
        ('--imdb', '--im'): {
            'help': 'IMDb ID or URL',
            'type': utils.argtypes.webdb_id('imdb'),
        },
        ('--source', '--so'): {
            'help': (
                'Original source of this release\n'
                + 'Should vaguely match: '
                + ', '.join(metadata.sources)
            ),
        },
        ('--type', '--ty'): {
            'help': (
                'General category of this release\n'
                + 'Must vaguely match: '
                + ', '.join(metadata.types)
            ),
            'type': PtpReleaseType,
        },
        ('--screenshots-count', '--ssc'): {
            'help': 'How many screenshots to make per video file',
            'type': utils.argtypes.number_of_screenshots(min=1, max=10),
        },
        ('--screenshots', '--ss'): {
            'help': (
                'Path(s) to existing screenshot file(s)\n'
                'Directories are searched recursively.\n'
                'More screenshots are created if necessary.'
            ),
            'nargs': '+',
            'action': 'extend',
            'type': utils.argtypes.files_with_extension('png'),
            'metavar': 'SCREENSHOT',
        },
        ('--upload-token', '--ut'): {
            'help': 'Upload token from staff',
        },
        ('--only-description', '--od'): {
            'help': 'Only generate description (do not submit)',
            'action': 'store_true',
            'group': 'generate-metadata',
        },
    },
}



from .. import btclients, imagehosts, utils
from .. import trackers as trackers_module
from .config import ConfigConfig


class Config(utils.config.ConfigBase):
    config: ConfigConfig = ConfigConfig()
    clients: btclients.BtclientsConfig = btclients.BtclientsConfig()
    imghosts: imagehosts.ImagehostsConfig = imagehosts.ImagehostsConfig()
    trackers: trackers_module.TrackersConfig = trackers_module.TrackersConfig()



""":class:`typing.Annotated` types to use for :class:`~.SubsectionBase` fields"""

import typing

import pydantic

from .. import types


def string(*, default, description, secret=False):
    """:class:`str` or :class:`pydantic.Secretstr`"""
    return typing.Annotated[
        pydantic.SecretStr if secret else str,
        pydantic.Field(
            default=default,
            description=description,
        ),
    ]


def boolean(*, default, description):
    """:class:`~.types.Bool`"""
    return typing.Annotated[
        types.Bool,
        pydantic.BeforeValidator(types.Bool),
        pydantic.Field(
            default=default,
            description=description,
        ),
    ]


def integer(*, default, description, min=None, max=None):
    """:class:`~.types.Integer`"""
    Integer = types.Integer(min=min, max=max)
    return typing.Annotated[
        Integer,
        pydantic.BeforeValidator(Integer),
        pydantic.Field(
            default=default,
            description=description,
        ),
    ]


def choice(*, default, options, description, empty_ok=False, case_sensitive=True):
    """:class:`~.types.Choice`"""
    Choice = types.Choice(options=options, empty_ok=empty_ok, case_sensitive=case_sensitive)
    return typing.Annotated[
        Choice,
        pydantic.BeforeValidator(Choice),
        pydantic.Field(
            default=default,
            description=description,
        ),
    ]


def custom(*, cls, default, description):
    """Custom type"""
    return typing.Annotated[
        cls,
        pydantic.BeforeValidator(cls),
        pydantic.Field(
            default=default,
            description=description,
        ),
    ]



"""
Upload images to image hosting service
"""

import functools

from .... import imagehosts, jobs, utils
from .base import CommandBase


class upload_images(CommandBase):
    """Upload images to image hosting service"""

    names = ('upload-images', 'ui')

    cli_arguments = {}

    subcommand_name = 'IMAGE_HOST'
    subcommands = {
        imagehost.name: {
            'description': imagehost.description,
            'cli': {
                # Default arguments for all image hosts
                'IMAGE': {
                    'nargs': '+',
                    'help': 'Path to image file',
                },
                ('--thumb-width', '-t'): {
                    'help': 'Thumbnail width in pixels',
                    'type': utils.argtypes.integer,
                    'default': None,
                },
                # Custom arguments defined by image host
                **imagehost.cli_arguments,
            },
        }
        for imagehost in imagehosts.imagehost_classes()
    }

    @functools.cached_property
    def imagehost_name(self):
        """Lower-case image host name"""
        return self.args.subcommand.lower()

    @functools.cached_property
    def imagehost_config(self):
        """
        Relevant section in image host configuration file combined with CLI
        arguments where CLI arguments take precedence unless their value is
        `None`
        """
        config = dict(self.config['imghosts'][self.imagehost_name])
        if self.args.thumb_width is not None:
            config['thumb_width'] = self.args.thumb_width
        return config

    @functools.cached_property
    def jobs(self):
        return (
            jobs.imagehost.ImagehostJob(
                home_directory=self.home_directory,
                cache_directory=self.cache_directory,
                ignore_cache=self.args.ignore_cache,
                image_paths=self.args.IMAGE,
                imagehosts=(
                    imagehosts.imagehost(
                        name=self.imagehost_name,
                        config=self.imagehost_config,
                    ),
                ),
            ),
        )



"""
Concrete :class:`~.TrackerConfigBase` subclass for ANT
"""

import base64

from ... import utils
from .. import base


class AntTrackerConfig(base.TrackerConfigBase):
    base_url: base.config.base_url(
        base64.b64decode('aHR0cHM6Ly9hbnRoZWxpb24ubWU=').decode('ascii')
    )

    apikey: base.config.apikey('')

    announce_url: base.config.announce_url('', autofetched=False)

    anonymous: base.config.anonymous('no')

    exclude: base.config.exclude(
        base.exclude_regexes.checksums,
        base.exclude_regexes.images,
        base.exclude_regexes.nfo,
        base.exclude_regexes.samples,
    )


cli_arguments = {
    'submit': {
        ('--anonymous', '--an'): {
            'help': 'Hide your username for this submission',
            'action': 'store_true',
            # This must be `None` so it doesn't override the "anonymous" value from the config file.
            # See CommandBase.get_options().
            'default': None,
        },
        ('--nfo',): {
            'help': 'Path to NFO file (supersedes any *.nfo file found in the release directory)',
        },
        ('--tmdb', '--tm'): {
            'help': 'TMDb ID or URL',
            'type': utils.argtypes.webdb_id('tmdb'),
        },
    },
}



import aiobtclientapi

from . import base, fields


class RtorrentBtclientConfig(base.BtclientConfig):
    name: fields.name = aiobtclientapi.RtorrentAPI.name
    label: fields.label = aiobtclientapi.RtorrentAPI.label
    client: fields.client = aiobtclientapi.RtorrentAPI.name

    url: fields.url(aiobtclientapi.RtorrentAPI)
    username: fields.username
    password: fields.password
    translate_path: fields.translate_path
    check_after_add: fields.check_after_add


class RtorrentBtclient(base.Btclient):
    Config = RtorrentBtclientConfig



from .base import Btclient, BtclientConfig, BtclientsConfig, GenericBtclientConfig
from .deluge import DelugeBtclient
from .qbittorrent import QbittorrentBtclient
from .rtorrent import RtorrentBtclient
from .transmission import TransmissionBtclient


def client_names():
    r"""Return sequence of valid :class:`~.Btclient.name`\ s"""
    return tuple(
        cls.name
        for cls in Btclient.__subclasses__()
    )



__project_name__ = 'upsies'
__description__ = 'Media metadata aggregator'
__homepage__ = 'https://upsies.readthedocs.io'
__changelog__ = 'https://codeberg.org/plotski/upsies/raw/branch/master/NEWS'
__version__ = '2025.04.21'



"""
API for image hosting services
"""

import pydantic

from .. import utils
from .base import ImagehostBase, ImagehostConfigBase
from .common import UploadedImage


def imagehost_classes():
    """Return list of :class:`.ImagehostBase` subclasses"""
    return utils.subclasses(ImagehostBase, utils.submodules(__package__))


def imagehost(name, config=None, cache_directory=None):
    """
    Create :class:`.ImagehostBase` instance

    :param str name: Name of the image hosting service. A subclass of
        :class:`.ImagehostBase` with the same :attr:`~.ImagehostBase.name` must
        exist in one of this package's submodules.
    :param config: User configuration passed to the subclass specified by `name`
    :param cache_directory: Where to cache URLs of uploaded images or `None` to
        use :attr:`~upsies.constants.DEFAULT_CACHE_DIRECTORY`.

    :raise ValueError: if no matching subclass can be found

    :return: :class:`.ImagehostBase` instance
    """
    for cls in imagehost_classes():
        if cls.name == name:
            return cls(
                config=config or {},
                cache_directory=cache_directory,
            )
    raise ValueError(f'Unsupported image hosting service: {name}')


def imagehost_names():
    """Return sequence of valid `name` arguments for :func:`imagehost`"""
    return sorted(utils.CaseInsensitiveString(cls.name) for cls in imagehost_classes())


def _imagehost_config_fields():
    """
    Return :class:`dict` that maps field names (:attr:`.ImagehostBase.name`) to
    `(<ImagehostConfigBase subclass>, <ImagehostConfigBase instance>)` tuples
    """
    clses = sorted(
        imagehost_classes(),
        key=lambda cls: cls.name,
    )
    return {
        cls.name: (cls.Config, cls.Config())
        for cls in clses
    }

ImagehostsConfig = pydantic.create_model(
    'ImagehostsConfig',
    **_imagehost_config_fields(),
    __base__=utils.config.SectionBase,
)



import aiobtclientapi
import aiobtclientrpc

from .. import errors, utils
from . import base, fields


class QbittorrentBtclientConfig(base.BtclientConfig):
    name: fields.name = aiobtclientapi.QbittorrentAPI.name
    label: fields.label = aiobtclientapi.QbittorrentAPI.label
    client: fields.client = aiobtclientapi.QbittorrentAPI.name

    url: fields.url(aiobtclientapi.QbittorrentAPI)
    username: fields.username
    password: fields.password
    translate_path: fields.translate_path
    check_after_add: fields.check_after_add
    category: utils.config.fields.string(
        default='',
        description='qBittorrent category to add torrents to',
    )


class QbittorrentBtclient(base.Btclient):
    Config = QbittorrentBtclientConfig

    async def add_torrent(self, *args, **kwargs):
        infohash = await super().add_torrent(*args, **kwargs)

        if 'category' in self._config:
            await self._set_category(infohash, self._config['category'])

        return infohash

    async def _set_category(self, infohash, category):
        try:
            await self._api.call(
                'torrents/setCategory',
                hashes=infohash,
                category=category,
            )
        except aiobtclientrpc.RPCError as e:
            if 'incorrect category name' in str(e).lower():
                raise errors.TorrentAddError(f'Unknown category: {category}') from e
            else:
                raise e



"""
Standardized configuration and setup howto
"""

from ... import __project_name__, btclients, constants, utils


class Howto:
    def __init__(self, tracker_cls):
        self._tracker_cls = tracker_cls
        self._section = -1

    def join(self, *sections):
        return '\n'.join(sections).strip()

    @property
    def current_section(self):
        return self._section

    @property
    def next_section(self):
        self._section += 1
        return self._section

    @property
    def introduction(self):
        return utils.string.evaluate_fstring(
            self.join(
                (
                    '{howto.next_section}. How To Read This Howto\n'
                    '\n'
                    '   {howto.current_section}.1 Words in ALL_CAPS_AND_WITH_UNDERSCORES are placeholders.\n'
                    '   {howto.current_section}.2 Everything after "$" is a terminal command.\n'
                ),
                (
                    '{howto.next_section}. Configuration Defaults (Optional)\n'
                    '\n'
                    '    If you prefer, you can write all default values at once and then edit\n'
                    '    them in your favorite $EDITOR.\n'
                    '\n'
                    '    $ {executable} set --dump\n'
                    '    $ $EDITOR {tildify(constants.TRACKERS_FILEPATH)}\n'
                    '    $ $EDITOR {tildify(constants.IMGHOSTS_FILEPATH)}\n'
                    '    $ $EDITOR {tildify(constants.CLIENTS_FILEPATH)}\n'
                    '    $ $EDITOR {tildify(constants.CONFIG_FILEPATH)}\n'
                ),
            ),
            howto=self,
            executable=__project_name__,
            constants=constants,
            tildify=utils.fs.tildify_path,
        )

    @property
    def screenshots(self):
        def imagehost_names(tracker):
            return ', '.join(
                tracker.TrackerConfig.defaults['image_host'].item_type.options
            )

        return utils.string.evaluate_fstring(
            self.join(
                (
                    '{howto.next_section}. Screenshots (Optional)\n'
                    '\n'
                    '   {howto.current_section}.1 Specify how many screenshots to make.\n'
                    '       $ {executable} set trackers.{tracker.name}.screenshots NUMBER_OF_SCREENSHOTS\n'
                ),
                (
                    '   {howto.current_section}.2 Specify where screenshots should be uploaded to.\n'
                    '       $ {executable} set trackers.{tracker.name}.image_host IMAGE_HOST1,IMAGE_HOST2,...\n'
                    '       If IMAGE_HOST1 is down, try IMAGE_HOST2 and so on.\n'
                    '       Supported services: {imagehost_names(tracker)}\n'
                    '\n'
                    '   {howto.current_section}.3 Configure image hosting service.\n'
                    '       This usually means configuring an API key, which may not be necessary.\n'
                    '       $ {executable} upload-images IMAGE_HOST --help\n'
                ),
            ),
            howto=self,
            tracker=self._tracker_cls,
            executable=__project_name__,
            imagehost_names=imagehost_names,
        )

    @property
    def autoseed(self):
        return utils.string.evaluate_fstring(
            self.join(
                (
                    '{howto.next_section}. Add Uploaded Torrents To BitTorrent Client (Optional)\n'
                    '\n'
                    '   {howto.current_section}.1 Specify at least one client connection.\n'
                    '\n'
                    '       $ {executable} set clients.CLIENT_NAME.client CLIENT\n'
                    '       $ {executable} set clients.CLIENT_NAME.url URL\n'
                    '       $ {executable} set clients.CLIENT_NAME.username USERNAME\n'
                    '       $ {executable} set clients.CLIENT_NAME.password PASSWORD\n'
                    '\n'
                    '       CLIENT must be one of: {client_names}\n'
                    '\n'
                    '       CLIENT_NAME can by any string (within reason) that identifies the client,\n'
                    '       e.g. "nas" or "seedbox". If CLIENT_NAME and CLIENT are identical, you don\'t\n'
                    '       have to set clients.CLIENT_NAME.client.\n'
                    '\n'
                    '   {howto.current_section}.2 Specify client to which uploaded torrents should be added to.\n'
                    '       $ {executable} set trackers.{tracker.name}.add_to CLIENT_NAME\n'
                ),
                (
                    '{howto.next_section}. Copy Uploaded Torrents To Directory (Optional)\n'
                    '\n'
                    '   $ {executable} set trackers.{tracker.name}.copy_to /path/to/directory\n'
                ),
            ),
            howto=self,
            tracker=self._tracker_cls,
            executable=__project_name__,
            client_names=', '.join(btclients.client_names()),
        )

    @property
    def reuse_torrents(self):
        return utils.string.evaluate_fstring(
            self.join(
                (
                    '{howto.next_section}. Reuse Existing Torrents (Optional)\n'
                    '\n'
                    '    You can skip the hashing when creating a torrent by specifying directory paths\n'
                    '    that contain the torrents you are seeding. A matching torrent is found by searching\n'
                    '    each directory recursively for a torrent with the same size and file names. If such\n'
                    '    a torrent is found, a few pieces of each file are hashed to verify the match.\n'
                    '\n'
                    '    $ {executable} set config.torrent-create.reuse_torrent_paths TORRENT_DIRECTORY1 TORRENT_DIRECTORY2 ...\n'
                ),
            ),
            howto=self,
            executable=__project_name__,
        )

    @property
    def upload(self):
        return utils.string.evaluate_fstring(
            self.join(
                (
                    '{howto.next_section}. Upload\n'
                    '\n'
                    '   $ {executable} submit {tracker.name} --help\n'
                    '   $ {executable} submit {tracker.name} /path/to/content\n'
                ),
            ),
            howto=self,
            tracker=self._tracker_cls,
            executable=__project_name__,
        )
